@@ -1,8 +1,8 @@
-/* Copyright (C) 2007 Eugeniy Meshcheryakov <eugen@debian.org>
+/* Copyright © 2007-2008 Євгеній Мещеряков <eugen@debian.org>
  *
- * This program is free software; you can redistribute it and/or modify
+ * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -11,8 +11,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 #include <ft2build.h>
 #include FT_FREETYPE_H
@@ -31,8 +30,14 @@
 #include <getopt.h>
 #include <stdint.h>
 #include <pango/pangocairo.h>
+#include <math.h>
+#include <libintl.h>
+#include <locale.h>
 
 #include "unicode_blocks.h"
+#include "config.h"
+
+#define _(str)	gettext(str)
 
 #define A4_WIDTH	(8.3*72)
 #define A4_HEIGHT	(11.7*72)
@@ -42,6 +47,9 @@
 #define cell_width	((A4_WIDTH - 2*xmin_border) / 16)
 #define cell_height	((A4_HEIGHT - 2*ymin_border) / 16)
 
+#define CELL_X(x_min, N)	((x_min) + cell_width * ((N) / 16))
+#define CELL_Y(N)	(ymin_border + cell_height * ((N) % 16))
+
 static struct option longopts[] = {
   {"font-file", 1, 0, 'f'},
   {"output-file", 1, 0, 'o'},
@@ -89,6 +97,10 @@ static PangoFontDescription *font_name_font;
 static PangoFontDescription *table_numbers_font;
 static PangoFontDescription *cell_numbers_font;
 
+static double cell_label_offset;
+static double cell_glyph_bot_offset;
+static double glyph_baseline_offset;
+
 static void usage(const char *);
 
 static struct fntsample_style *find_style(const char *name)
@@ -120,14 +132,14 @@ static int set_style(const char *name, const char *val)
 	if (style->val)
 		free(style->val);
 	style->val = new_val;
-	
+
 	return 0;
 }
 
 static const char *get_style(const char *name)
 {
 	struct fntsample_style *style = find_style(name);
-	
+
 	if (!style)
 		return NULL;
 
@@ -145,6 +157,11 @@ static int parse_style_string(char *s)
 	return set_style(s, n);
 }
 
+/*
+ * Update output range.
+ *
+ * Returns -1 on error.
+ */
 static int add_range(char *range, bool include)
 {
 	struct range *r;
@@ -199,6 +216,10 @@ static int add_range(char *range, bool include)
 	return 0;
 }
 
+/*
+ * Check if character with the given code belongs
+ * to output range specified by the user.
+ */
 static bool in_range(uint32_t c)
 {
 	bool in = ranges ? (!ranges->include) : 1;
@@ -211,6 +232,13 @@ static bool in_range(uint32_t c)
 	return in;
 }
 
+/*
+ * Get glyph index for the next glyph from the given font face, that
+ * represents character from output range specified by the user.
+ *
+ * Returns character code, updates 'idx'.
+ * 'idx' can became 0 if there are no more glyphs.
+ */
 static FT_ULong get_next_char(FT_Face face, FT_ULong charcode, FT_UInt *idx)
 {
 	FT_ULong rval = charcode;
@@ -222,6 +250,13 @@ static FT_ULong get_next_char(FT_Face face, FT_ULong charcode, FT_UInt *idx)
 	return rval;
 }
 
+/*
+ * Locate first character from the given font face that belongs
+ * to the user-specified output range.
+ *
+ * Returns character code, updates 'idx' with glyph index.
+ * Glyph index can became 0 if there are no matching glyphs in the font.
+ */
 static FT_ULong get_first_char(FT_Face face, FT_UInt *idx)
 {
 	FT_ULong rval;
@@ -234,7 +269,12 @@ static FT_ULong get_first_char(FT_Face face, FT_UInt *idx)
 	return rval;
 }
 
-static PangoLayout *draw_text(cairo_t *cr, PangoFontDescription *ftdesc, const char *text, PangoRectangle *r)
+/*
+ * Create Pango layout for the given text.
+ * Updates 'r' with text extents.
+ * Returned layout should be freed using g_object_unref().
+ */
+static PangoLayout *layout_text(cairo_t *cr, PangoFontDescription *ftdesc, const char *text, PangoRectangle *r)
 {
 	PangoLayout *layout;
 
@@ -259,14 +299,14 @@ static void parse_options(int argc, char * const argv[])
 		switch (c) {
 		case 'f':
 			if (font_file_name) {
-				fprintf(stderr, "Font file name should be given only once!\n");
+				fprintf(stderr, _("Font file name should be given only once!\n"));
 				exit(1);
 			}
 			font_file_name = optarg;
 			break;
 		case 'o':
 			if (output_file_name) {
-				fprintf(stderr, "Output file name should be given only once!\n");
+				fprintf(stderr, _("Output file name should be given only once!\n"));
 				exit(1);
 			}
 			output_file_name = optarg;
@@ -277,7 +317,7 @@ static void parse_options(int argc, char * const argv[])
 			break;
 		case 'd':
 			if (other_font_file_name) {
-				fprintf(stderr, "Font file name should be given only once!\n");
+				fprintf(stderr, _("Font file name should be given only once!\n"));
 				exit(1);
 			}
 			other_font_file_name = optarg;
@@ -314,6 +354,10 @@ static void parse_options(int argc, char * const argv[])
 	}
 }
 
+/*
+ * Locate unicode block that contains given character code.
+ * Returns this block or NULL if not found.
+ */
 static const struct unicode_block *get_unicode_block(unsigned long charcode)
 {
 	const struct unicode_block *block;
@@ -325,53 +369,76 @@ static const struct unicode_block *get_unicode_block(unsigned long charcode)
 	return NULL;
 }
 
-static int is_in_block(unsigned long charcode, const struct unicode_block *block)
+/*
+ * Check if the given character code belongs to the given Unicode block.
+ */
+static bool is_in_block(unsigned long charcode, const struct unicode_block *block)
 {
 	return ((charcode >= block->start) && (charcode <= block->end));
 }
 
+/*
+ * Format and print outline information, if requested by the user.
+ */
 static void outline(int level, int page, const char *text)
 {
 	if (print_outline)
 		printf("%d %d %s\n", level, page, text);
 }
 
-static void draw_header(cairo_t *cr, const char *face_name, const char *range_name)
+/*
+ * Draw header of a page.
+ * Header shows font name and current Unicode block.
+ */
+static void draw_header(cairo_t *cr, const char *face_name, const char *block_name)
 {
 	PangoLayout *layout;
 	PangoRectangle r;
 
-	layout = draw_text(cr, font_name_font, face_name, &r);
+	layout = layout_text(cr, font_name_font, face_name, &r);
 	cairo_move_to(cr, (A4_WIDTH - (double)r.width/PANGO_SCALE)/2.0, 30.0);
 	pango_cairo_show_layout_line(cr, pango_layout_get_line_readonly(layout, 0));
 	g_object_unref(layout);
 
-	layout = draw_text(cr, header_font, range_name, &r);
+	layout = layout_text(cr, header_font, block_name, &r);
 	cairo_move_to(cr, (A4_WIDTH - (double)r.width/PANGO_SCALE)/2.0, 50.0);
 	pango_cairo_show_layout_line(cr, pango_layout_get_line_readonly(layout, 0));
 	g_object_unref(layout);
 }
 
-static void draw_cell(cairo_t *cr, double x, double y,
-		unsigned long idx, bool highlight, cairo_glyph_t *glyph)
+/*
+ * Highlight the cell with given coordinates.
+ * Used to highlight new glyphs.
+ */
+static void highlight_cell(cairo_t *cr, double x, double y)
+{
+	cairo_save(cr);
+	cairo_set_source_rgb(cr, 1.0, 1.0, 0.6);
+	cairo_rectangle(cr, x, y, cell_width, cell_height);
+	cairo_fill(cr);
+	cairo_restore(cr);
+}
+
+/*
+ * Try to place glyph with the given index at the middle of the cell.
+ * Changes argument 'glyph'
+ */
+static void position_glyph(cairo_t *cr, double x, double y,
+		unsigned long idx, cairo_glyph_t *glyph)
 {
 	cairo_text_extents_t extents;
 
-	if (highlight) {
-		cairo_set_source_rgb(cr, 1.0, 1.0, 0.6);
-		cairo_rectangle(cr, x, y, cell_width, cell_height);
-		cairo_fill(cr);
-		cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
-	}
-
 	*glyph = (cairo_glyph_t){idx, 0, 0};
 
 	cairo_glyph_extents(cr, glyph, 1, &extents);
 
 	glyph->x += x + (cell_width - extents.width)/2.0 - extents.x_bearing;
-	glyph->y += y + cell_height / 2.0;
+	glyph->y += y + glyph_baseline_offset;
 }
 
+/*
+ * Draw table grid with row and column numbers.
+ */
 static void draw_grid(cairo_t *cr, unsigned int x_cells,
 		unsigned long block_start)
 {
@@ -411,7 +478,7 @@ static void draw_grid(cairo_t *cr, unsigned int x_cells,
 
 	for (i = 0; i < 16; i++) {
 		buf[0] = hexdigs[i];
-		layout = draw_text(cr, table_numbers_font, buf, &r);
+		layout = layout_text(cr, table_numbers_font, buf, &r);
 		cairo_move_to(cr, x_min - (double)PANGO_RBEARING(r)/PANGO_SCALE - 5.0,
 				72.0 + (i+0.5) * TABLE_H/16 + (double)PANGO_DESCENT(r)/PANGO_SCALE/2);
 		pango_cairo_show_layout_line(cr, pango_layout_get_line_readonly(layout, 0));
@@ -423,7 +490,7 @@ static void draw_grid(cairo_t *cr, unsigned int x_cells,
 
 	for (i = 0; i < x_cells; i++) {
 		snprintf(buf, sizeof(buf), "%03lX", block_start / 16 + i);
-		layout = draw_text(cr, table_numbers_font, buf, &r);
+		layout = layout_text(cr, table_numbers_font, buf, &r);
 		cairo_move_to(cr, x_min + i*cell_width + (cell_width - (double)r.width/PANGO_SCALE)/2,
 				ymin_border - 5.0);
 		pango_cairo_show_layout_line(cr, pango_layout_get_line_readonly(layout, 0));
@@ -431,8 +498,12 @@ static void draw_grid(cairo_t *cr, unsigned int x_cells,
 	}
 }
 
-static void draw_empty_cell(cairo_t *cr, double x, double y, unsigned long charcode)
+/*
+ * Fill empty cell. Color of the fill depends on the character properties.
+ */
+static void fill_empty_cell(cairo_t *cr, double x, double y, unsigned long charcode)
 {
+	cairo_save(cr);
 	if (g_unichar_isdefined(charcode)) {
 		if (g_unichar_iscntrl(charcode))
 			cairo_set_source_rgb(cr, 0.0, 0.0, 0.5);
@@ -441,10 +512,12 @@ static void draw_empty_cell(cairo_t *cr, double x, double y, unsigned long charc
 	}
 	cairo_rectangle(cr, x, y, cell_width, cell_height);
 	cairo_fill(cr);
-	if (g_unichar_isdefined(charcode))
-		cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
+	cairo_restore(cr);
 }
 
+/*
+ * Draw label with character code.
+ */
 static void draw_charcode(cairo_t *cr, double x, double y, FT_ULong charcode)
 {
 	char buf[9];
@@ -452,13 +525,22 @@ static void draw_charcode(cairo_t *cr, double x, double y, FT_ULong charcode)
 	PangoRectangle r;
 
 	snprintf(buf, sizeof(buf), "%04lX", charcode);
-	layout = draw_text(cr, cell_numbers_font, buf, &r);
-	cairo_move_to(cr, x + (cell_width - (double)r.width/PANGO_SCALE)/2.0, y + cell_height - 4.0);
+	layout = layout_text(cr, cell_numbers_font, buf, &r);
+	cairo_move_to(cr, x + (cell_width - (double)r.width/PANGO_SCALE)/2.0, y + cell_height - cell_label_offset);
 	pango_cairo_show_layout_line(cr, pango_layout_get_line_readonly(layout, 0));
 	g_object_unref(layout);
 }
 
-static int draw_unicode_block(cairo_t *cr, cairo_font_face_t *face,
+/*
+ * Draws tables for all characters in the given Unicode block.
+ * Use font described by face and ft_face. Start from character
+ * with given charcode (it should belong to the given Unicode
+ * block). After return 'charcode' equals the last character code
+ * of the block.
+ *
+ * Returns number of pages drawn.
+ */
+static int draw_unicode_block(cairo_t *cr, cairo_scaled_font_t *font,
 		FT_Face ft_face, const char *fontname, unsigned long *charcode,
 		const struct unicode_block *block, FT_Face ft_other_face)
 {
@@ -478,7 +560,6 @@ static int draw_unicode_block(cairo_t *cr, cairo_font_face_t *face,
 		double x_min = (A4_WIDTH - rows * cell_width) / 2;
 		unsigned long i;
 		bool filled_cells[256]; /* 16x16 glyphs max */
-		bool highlight = false;
 
 		/* XXX WARNING: not reentrant! */
 		static cairo_glyph_t glyphs[256];
@@ -490,39 +571,50 @@ static int draw_unicode_block(cairo_t *cr, cairo_font_face_t *face,
 
 		memset(filled_cells, '\0', sizeof(filled_cells));
 
-		cairo_set_font_face(cr, face);
-		cairo_set_font_size(cr, 20.0);
+		cairo_set_scaled_font(cr, font);
+		/*
+		 * Fill empty cells and calculate coordinates of the glyphs.
+		 * Also highlight cells if needed.
+		 */
 		do {
+			/* the current glyph position in the table */
+			int charpos = *charcode - tbl_start;
+
+			/* fill empty cells before the current glyph */
 			for (i = prev_cell + 1; i < *charcode; i++) {
-				draw_empty_cell(cr, x_min + cell_width*((i - tbl_start) / 16),
-							ymin_border + cell_height*((i - tbl_start) % 16), i);
+				int pos = i - tbl_start;
+				fill_empty_cell(cr, CELL_X(x_min, pos), CELL_Y(pos), i);
 			}
 
-			if (ft_other_face)
-				highlight = !FT_Get_Char_Index(ft_other_face, *charcode);
+			/* if it is new glyph - highlight the cell */
+			if (ft_other_face && !FT_Get_Char_Index(ft_other_face, *charcode))
+				highlight_cell(cr, CELL_X(x_min, charpos), CELL_Y(charpos));
 
-			draw_cell(cr, x_min + cell_width*((*charcode - tbl_start) / 16),
-					ymin_border + cell_height*((*charcode - tbl_start) % 16),
-					idx, highlight, &glyphs[nglyphs++]);
+			/* For now just position glyphs. They will be shown later,
+			 * to make output more efficient. */
+			position_glyph(cr, CELL_X(x_min, charpos), CELL_Y(charpos),
+					idx, &glyphs[nglyphs++]);
 
-			filled_cells[*charcode - tbl_start] = true;
+			filled_cells[charpos] = true;
 
 			prev_charcode = *charcode;
 			prev_cell = *charcode;
 			*charcode = get_next_char(ft_face, *charcode, &idx);
 		} while (idx && (*charcode < tbl_end) && is_in_block(*charcode, block));
 
+		/* Fill remaining empty cells */
+		for (i = prev_cell + 1; i < tbl_end; i++) {
+			int pos = i - tbl_start;
+			fill_empty_cell(cr, CELL_X(x_min, pos), CELL_Y(pos), i);
+		}
+
+		/* Show previously positioned glyphs */
 		cairo_show_glyphs(cr, glyphs, nglyphs);
 
 		for (i = 0; i < tbl_end - tbl_start; i++)
 			if (filled_cells[i])
-				draw_charcode(cr, x_min + cell_width*(i / 16), ymin_border + cell_height*(i % 16),
+				draw_charcode(cr, CELL_X(x_min, i), CELL_Y(i),
 						i + tbl_start);
-		for (i = prev_cell + 1; i < tbl_end; i++) {
-			draw_empty_cell(cr, x_min + cell_width*((i - tbl_start) / 16),
-					ymin_border + cell_height*((i - tbl_start) % 16),
-					i);
-		}
 
 		draw_grid(cr, rows, tbl_start);
 		npages++;
@@ -534,7 +626,10 @@ static int draw_unicode_block(cairo_t *cr, cairo_font_face_t *face,
 	return npages;
 }
 
-static void draw_glyphs(cairo_t *cr, cairo_font_face_t *face, FT_Face ft_face,
+/*
+ * The main drawing function.
+ */
+static void draw_glyphs(cairo_t *cr, cairo_scaled_font_t *font, FT_Face ft_face,
 		const char *fontname, FT_Face ft_other_face)
 {
 	FT_ULong charcode;
@@ -551,7 +646,7 @@ static void draw_glyphs(cairo_t *cr, cairo_font_face_t *face, FT_Face ft_face,
 		if (block) {
 			int npages;
 			outline(1, pageno, block->name);
-			npages = draw_unicode_block(cr, face, ft_face, fontname,
+			npages = draw_unicode_block(cr, font, ft_face, fontname,
 					&charcode, block, ft_other_face);
 			pageno += npages;
 		}
@@ -559,13 +654,16 @@ static void draw_glyphs(cairo_t *cr, cairo_font_face_t *face, FT_Face ft_face,
 	}
 }
 
+/*
+ * Print usage instructions and default values for styles
+ */
 static void usage(const char *cmd)
 {
 	const struct fntsample_style *style;
 
-	fprintf(stderr, "Usage: %s [ OPTIONS ] -f FONT-FILE -o OUTPUT-FILE\n"
-			"       %s -h\n\n" , cmd, cmd);
-	fprintf(stderr, "Options:\n"
+	fprintf(stderr, _("Usage: %s [ OPTIONS ] -f FONT-FILE -o OUTPUT-FILE\n"
+			"       %s -h\n\n") , cmd, cmd);
+	fprintf(stderr, _("Options:\n"
 			"  --font-file,         -f FONT-FILE    Create samples of FONT-FILE\n"
 			"  --output-file,       -o OUTPUT-FILE  Save samples to OUTPUT-FILE\n"
 			"  --help,              -h              Show this information message and exit\n"
@@ -574,12 +672,17 @@ static void usage(const char *cmd)
 			"  --print-outline,     -l              Print document outlines data to standard output\n"
 			"  --include-range,     -i RANGE        Show characters in RANGE\n"
 			"  --exclude-range,     -x RANGE        Do not show characters in RANGE\n"
-			"  --style,             -t \"STYLE: VAL\" Set STYLE to value VAL\n");
-	fprintf(stderr, "\nSupported styles (and default values):\n");
+			"  --style,             -t \"STYLE: VAL\" Set STYLE to value VAL\n"));
+	fprintf(stderr, _("\nSupported styles (and default values):\n"));
 	for (style = styles; style->name; style++)
 		fprintf(stderr, "\t%s (%s)\n", style->name, style->default_val);
 }
 
+/*
+ * Try to get font name for a given font face.
+ * Returned name should be free()'d after use.
+ * If function cannot allocate memory, it terminates the program.
+ */
 static const char *get_font_name(FT_Face face)
 {
 	FT_Error error;
@@ -629,6 +732,9 @@ static const char *get_font_name(FT_Face face)
 	return fontname;
 }
 
+/*
+ * Initialize fonts used to print table heders and character codes.
+ */
 static void init_pango_fonts(void)
 {
 	/* FIXME is this correct? */
@@ -642,6 +748,71 @@ static void init_pango_fonts(void)
 	cell_numbers_font = pango_font_description_from_string(get_style("cell-numbers-font"));
 }
 
+/*
+ * Calculate various offsets.
+ */
+static void calculate_offsets(cairo_t *cr)
+{
+	PangoRectangle extents;
+	/* Assume that vertical extents does not depend on actual text */
+	PangoLayout *l = layout_text(cr, cell_numbers_font, "0123456789ABCDEF", &extents);
+	g_object_unref(l);
+	/* Unsolved mistery of pango's font metrics.... */
+	double digits_ascent = pango_units_to_double(PANGO_DESCENT(extents));
+	double digits_descent = -pango_units_to_double(PANGO_ASCENT(extents));
+
+	cell_label_offset = digits_descent + 2;
+	cell_glyph_bot_offset = cell_label_offset + digits_ascent + 2;
+}
+
+/*
+ * Create cairo scaled font with the best size (hopefuly...)
+ */
+static cairo_scaled_font_t *create_default_font(FT_Face ft_face)
+{
+	cairo_font_face_t *cr_face = cairo_ft_font_face_create_for_ft_face(ft_face, 0);
+	cairo_matrix_t font_matrix;
+	cairo_matrix_t ctm;
+	cairo_font_options_t *options = cairo_font_options_create();
+	cairo_scaled_font_t *cr_font;
+	cairo_font_extents_t extents;
+
+
+	/* First create font with size 1 and measure it */
+	cairo_matrix_init_identity(&font_matrix);
+	cairo_matrix_init_identity(&ctm);
+	/* Turn off rounding, so we can get real metrics */
+	cairo_font_options_set_hint_metrics(options, CAIRO_HINT_METRICS_OFF);
+	cr_font = cairo_scaled_font_create(cr_face, &font_matrix, &ctm, options);
+	cairo_scaled_font_extents(cr_font, &extents);
+
+	/* Use some magic to find the best font size... */
+	double tgt_size = cell_height - cell_glyph_bot_offset - 2;
+	if (tgt_size <= 0) {
+		fprintf(stderr, _("Not enough space for rendering glyphs. Make cell font smaller.\n"));
+		exit(5);
+	}
+	double act_size = extents.ascent + extents.descent;
+	if (act_size <= 0) {
+		fprintf(stderr, _("The font has strange metrics: ascent + descent = %g\n"), act_size);
+		exit(5);
+	}
+	double scale = tgt_size / act_size;
+	if (scale > 1)
+		scale = trunc(scale); // just to make numbers nicer
+	if (scale > 20)
+		scale = 20; // Do not make font larger than in previous versions
+
+	cairo_scaled_font_destroy(cr_font);
+
+	/* Create the font once again, but this time scaled */
+	cairo_matrix_init_scale(&font_matrix, scale, scale);
+	cr_font = cairo_scaled_font_create(cr_face, &font_matrix, &ctm, options);
+	cairo_scaled_font_extents(cr_font, &extents);
+	glyph_baseline_offset = (tgt_size - (extents.ascent + extents.descent)) / 2 + 2 + extents.ascent;
+	return cr_font;
+}
+
 int main(int argc, char **argv)
 {
 	cairo_surface_t *surface;
@@ -650,31 +821,34 @@ int main(int argc, char **argv)
 	FT_Library library;
 	FT_Face face, other_face = NULL;
 	const char *fontname; /* full name of the font */
-	cairo_font_face_t *cr_face;
 	cairo_status_t cr_status;
+	cairo_scaled_font_t *cr_font;
+
+	setlocale(LC_ALL, "");
+	bindtextdomain(PACKAGE, LOCALEDIR);
+	textdomain(PACKAGE);
 
 	parse_options(argc, argv);
 
 	error = FT_Init_FreeType(&library);
 	if (error) {
-		fprintf(stderr, "%s: freetype error\n", argv[0]);
+		/* TRANSLATORS: 'freetype' is a name of a library, and should be left untranslated */
+		fprintf(stderr, _("%s: freetype error\n"), argv[0]);
 		exit(3);
 	}
 
 	error = FT_New_Face(library, font_file_name, 0, &face);
 	if (error) {
-		fprintf(stderr, "%s: failed to create new face\n", argv[0]);
+		fprintf(stderr, _("%s: failed to open font file %s\n"), argv[0], font_file_name);
 		exit(4);
 	}
 
 	fontname = get_font_name(face);
 
-	cr_face = cairo_ft_font_face_create_for_ft_face(face, 0);
-
 	if (other_font_file_name) {
 		error = FT_New_Face(library, other_font_file_name, 0, &other_face);
 		if (error) {
-			fprintf(stderr, "%s: failed to create new face\n", argv[0]);
+			fprintf(stderr, _("%s: failed to create new font face\n"), argv[0]);
 			exit(4);
 		}
 	}
@@ -686,7 +860,8 @@ int main(int argc, char **argv)
 
 	cr_status = cairo_surface_status(surface);
 	if (cr_status != CAIRO_STATUS_SUCCESS) {
-		fprintf(stderr, "%s: failed to create cairo surface: %s\n",
+		/* TRANSLATORS: 'cairo' is a name of a library, and should be left untranslated */
+		fprintf(stderr, _("%s: failed to create cairo surface: %s\n"),
 				argv[0], cairo_status_to_string(cr_status));
 		exit(1);
 	}
@@ -694,7 +869,7 @@ int main(int argc, char **argv)
 	cr = cairo_create(surface);
 	cr_status = cairo_status(cr);
 	if (cr_status != CAIRO_STATUS_SUCCESS) {
-		fprintf(stderr, "%s: cairo_create failed: %s\n",
+		fprintf(stderr, _("%s: cairo_create failed: %s\n"),
 				argv[0], cairo_status_to_string(cr_status));
 		exit(1);
 	}
@@ -702,9 +877,17 @@ int main(int argc, char **argv)
 	cairo_surface_destroy(surface);
 
 	init_pango_fonts();
+	calculate_offsets(cr);
+	cr_font = create_default_font(face);
+	cr_status = cairo_scaled_font_status(cr_font);
+	if (cr_status != CAIRO_STATUS_SUCCESS) {
+		fprintf(stderr, _("%s: failed to create scaled font: %s\n"),
+				argv[0], cairo_status_to_string(cr_status));
+		exit(1);
+	}
 
 	cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
-	draw_glyphs(cr, cr_face, face, fontname, other_face);
+	draw_glyphs(cr, cr_font, face, fontname, other_face);
 	cairo_destroy(cr);
 	return 0;
 }
